import Foundation

import Model
import CustomTypes
import Persistance

var game: Game? = nil

enum EnumRules: String, CaseIterable {
    case FourInARow = "Four in a row"
    case TicTacToe = "Tic Tac Toe"
}

func enumPrompt<E: RawRepresentable<String> & CaseIterable>(prompt: String) -> E where E.AllCases.Index == Int {
    print("\(prompt):")
    E.allCases.enumerated().forEach { (index, variant) in
        print("\t\(index + 1). \(variant.rawValue)")
    }
    
    guard let rawInput = readLine(strippingNewline: true) else { exit(EXIT_SUCCESS) }
    guard let index = Int(rawInput), index >= 1, index <= E.allCases.count
    else {
        fatalError("Invalid variant number")
    }
    
    return E.allCases[index - 1]
}

enum PlayerType: String, CaseIterable {
    case AIRandom = "(AI) Random"
    case Human = "Player"
}

@MainActor
func playerPrompt(allowed_moves: [Move.Action], board: Board) async -> Move.Action {
    for (i, action) in allowed_moves.enumerated() {
        let text = switch action {
        case .InsertAt(let at): "Place piece at \(at.col):\(at.row)"
        case .InsertOnSide(let side, let offset): "Fall piece from \(side), offset \(offset)"
        }
        print("\(i + 1). \(text)")
    }
    
    if let rawInput = readLine(strippingNewline: true) {
        guard let index = Int(rawInput), index >= 1, index <= allowed_moves.count
        else {
            fatalError("Invalid move number")
        }
        
        return allowed_moves[index - 1]
    } else {
        // dump game to disk
        
        do {
            try await Storage.save(game: game!, finished: false)
        } catch {
            fatalError("Failed to save game: \(error)")
        }
        
        exit(EXIT_SUCCESS)
    }
}

func createPlayer(type: PlayerType, playing pieces: PieceType, named name: String) -> Player {
    switch type {
    case .Human:
        HumanPlayer(name: name, piece_type: pieces, callback: playerPrompt)
    case .AIRandom:
        RandomPlayer(name: name, piece_type: pieces)
    }
}

if let unfinished = try await Storage.loadUnfinishedGame(callbackFactory: { name, type in playerPrompt }) {
    print("There was an unfinished game. Resume? [Y/n]")
    
    guard let answer = readLine(strippingNewline: true) else {
        exit(EXIT_SUCCESS)
    }
    
    switch answer.lowercased() {
    case "", "y", "yes":
        game = unfinished
    default:
        break;
    }
    
    try await Storage.clearUnfinished()
}

if game == nil {
    let gameType: EnumRules = enumPrompt(prompt: "Choose a game type")

    let players: [Player] = PieceType.allCases.map { type in
        let ptype: PlayerType = enumPrompt(prompt: "Player playing pieces \(type)")
        
        print("Enter player name:")
        guard let playerName = readLine(strippingNewline: true) else {
            fatalError("Player name unspecified")
        }
        
        return createPlayer(type: ptype, playing: type, named: playerName)
    }

    let rules: any Rules = switch gameType {
    case .FourInARow: FourInARowRules(players: players)!
    case .TicTacToe: TicTacToeRules(players: players)!
    }
    
    game = Game(players: players, rules: rules)
}

game!.boardChanged.add { board in
    print(board)
}
game!.moveIsInvalid.add { move in
    print("\(move.player.name), this move is invalid according to the rules.")
}
game!.gameStateChanged.add { state in
    switch state {
    case .Playing(let player):
        print("\(player.name) turn.")
    case .Win(let winner, let board, let cells):
        print("\(winner.name) is the winner")
        print(board.display(winCells: cells))
    case .Draw:
        print("Its a daw.")
    case .Invalid:
        fatalError("Game is invalid!")
    }
}

await game!.play()

try await Storage.save(game: game!, finished: true)
