public struct FourInARowRules: Rules {
    public static let COLUMNS_MIN: Int = 3
    public static let COLUMNS_DEFAULT: Int = 7
    public static let ROWS_MIN: Int = 3
    public static let ROWS_DEFAULT: Int = 6
    
    private let players: [Player]

    // // internal for unit testing purposes
    // internal(set) public var state: GameState

    // private(set) public var history: [Move] = []

    public let columns: Int, rows: Int, minAligned: Int
    
    public var type: RulesTypes = .FourInARow
    
    public init?(players: [Player]) {
        self.init(columns: Self.COLUMNS_DEFAULT, rows: Self.ROWS_DEFAULT, players: players)
    }

    public init?(columns: Int, rows: Int, minAligned: Int = 4, players: [Player]) {
        guard
            columns >= Self.COLUMNS_MIN,
            rows >= Self.ROWS_MIN,
            minAligned > 1,
            players.count >= 2
        else {
            return nil
        }

        self.players = players
        self.columns = columns
        self.rows = rows
        self.minAligned = minAligned
    }

    public func createBoard() -> Board {
        Board(columns: self.columns, rows: self.rows)!
    }

    public func isValid(board: Board, move: Move) -> Bool {
        guard self.isValid(board: board) else { return false }

        if case .InsertOnSide(.Top, let offset) = move.action, offset >= 0 && offset < self.columns {
            return board.fallCoordinates(
                initialCoords: Coords(offset, 0),
                direction: .Bottom
            ) != .Occupied
        }

        return false
    }

    public func isValid(board: Board) -> Bool {
        for c in 0..<board.columns {
            var had = false;

            for r in 0..<board.rows {
                switch board[c, r] != nil {
                case false where had:
                    return false

                case let has:
                    had = has
                }
            }
        }

        return true
    }

    public func validMoves(board: Board) -> [Move] {
        return self.players.flatMap({
            player in self.validMoves(board: board, for_player: player).map({
                action in Move(player: player, action: action)
            })
        })
    }

    public func validMoves(board: Board, for_player player: Player) -> [Move.Action] {
        var moves: [Move.Action] = [];

        for c in 0..<board.columns {
            switch board.fallCoordinates(initialCoords: Coords(c, 0), direction: .Bottom) {
            case .Border, .Piece:
                moves.append(.InsertOnSide(side: .Top, offset: c))

            default:
                break
            }
        }

        return moves
    }
    
    public func gameState(board: Board, last_turn: Player?) -> GameState {
        Self.rowOfLengthWin(board: board, last_turn: last_turn, minimum_aligned: self.minAligned, players: self.players)
    }
    
    internal static func rowOfLengthWin(board: Board, last_turn: Player?, minimum_aligned: Int, players: [Player]) -> GameState {
        var occupied = 0
        
        for column in 0..<board.columns {
            for row in 0..<board.rows {
                let current = Coords(column, row)
                guard let piece = board[current] else { continue }
                
                occupied += 1
                
                // For each "axis" (described as one direction)
                for dir: (Int, Int) in [(1, 0), (0, 1), (1, 1), (-1, 1)] {
                    let cells = Self.row(center: current, board: board, dir: dir)
                    
                    if cells.count >= minimum_aligned {
                        let player = players.first(where: { $0.piece_type == piece.type })!
                        return GameState.Win(winner: player, board: board, cells: cells)
                    }
                }
            }
        }
        
        if occupied == board.columns * board.rows {
            return .Draw
        } else if let last_turn = last_turn {
            return .Playing(turn: players[(players.firstIndex(of: last_turn)! + 1) % players.count])
        } else {
            return .Playing(turn: players.first!)
        }
    }

    /* public mutating func onMoveDone(move: Move, board: Board) -> Void {
        // self.history.append(move)

        switch move.action {
        case .InsertOnSide(side: .Top, let offset):
            let initCoords = board.getInsertionCoordinates(from: .Top, offset: offset)
            let pieceCoords = switch board.fallCoordinates(
                initialCoords: initCoords,
                direction: .Bottom
            ) {
            case .Occupied:
                initCoords
            case .Piece(_, let touched):
                touched
                
            default:
                fatalError("Illegal move \(move.action)")
            }
            
            if Self.countMaxRow(center: pieceCoords, board: board) >= self.minAligned {
                self.state = .Finished(winner: move.player)
            } else if board.countPieces() == board.columns * board.rows {
                self.state = .Finished(winner: nil)
            } else {
                let next: PieceType = switch move.player {
                case .A: .B
                case .B: .A
                }
                self.state = .Playing(turn: next)
            }
        default:
            fatalError("Illegal move \(move.action)")
        }
    } */
    
    @available(*, deprecated, message: "Old Rules design remnents")
    internal static func countMaxRow(center: Coords, board: Board) -> Int {
        var maxLength = 0
        // For each "axis" (described as one direction)
        for dir: (dc: Int, dr: Int) in [(1, 0), (0, 1), (1, 1), (-1, 1)] {
            maxLength = max(maxLength, row(center: center, board: board, dir: dir).count)
        }
        
        return maxLength
    }
    
    private static func row(center: Coords, board: Board, dir: (dc: Int, dr: Int)) -> [Coords] {
        guard let of = board[center]?.type else { return [] }
        
        var cells = [center]
        
        // Run in the two opposite directions of the axis
        for (dc, dr) in [(dir.dc, dir.dr), (-dir.dc, -dir.dr)] {
            var pos = center
            
            while true {
                pos = Coords(pos.col + dc, pos.row + dr)
                if !board.isInBounds(pos) || board[pos]?.type != of { break }
                cells.append(pos)
            }
        }
        
        return cells
    }
}
