import Model

/// This struct is a workaround as it seems codeable required initializer to be required, which is not possible from an extension
public struct CodableGameWrapper: Codable {
    public let game: Game
    
    public init(of game: Game) {
        self.game = game
    }
    
    enum CodingKeys: CodingKey {
        case players, rules, board, state
    }
    
    public init(from decoder: any Decoder) throws {
        guard let context = decoder.userInfo[.gameDecodingContext] as? CodableContext
        else { throw DeError.CodableContext }
        
        let container = try decoder.container(keyedBy: CodingKeys.self)
        
        context.players = (try container.decode([CodablePlayerWrapper].self, forKey: .players)).map { $0.player }
        
        context.board = try container.decode(Board.self, forKey: .board)
        
        let rules = (try container.decode(CodableRulesWrapper.self, forKey: .rules)).rules
        
        let state = try container.decode(GameState.self, forKey: .state)
        
        guard let game = Game.init(players: context.players!, rules: rules, board: context.board!, state: state)
        else { throw DeError.InitializationFailed }
        
        self.game = game
    }
    
    public func encode(to encoder: any Encoder) throws {
        guard let context = encoder.userInfo[.gameDecodingContext] as? CodableContext
        else { throw DeError.CodableContext }
        
        var container = encoder.container(keyedBy: CodingKeys.self)
        
        context.players = self.game.players
        context.board = self.game.board
        
        try container.encode(self.game.players.map { CodablePlayerWrapper(of: $0) }, forKey: .players)
        
        try container.encode(self.game.board, forKey: .board)
        
        try container.encode(CodableRulesWrapper(of: self.game.rules), forKey: .rules)
        
        try container.encode(self.game.state, forKey: .state)
    }
}

extension GameState: Codable {
    enum VariantCodingKeys: CaseIterable, CodingKey {
        case Playing, Win, Draw, Invalid
    }
    
    public init(from decoder: any Decoder) throws {
        let container = try decoder.container(keyedBy: EnumCommonCodingKeys.self)
        
        self = switch (try container.decode(CodableEnum<VariantCodingKeys>.self, forKey: .type)).variant {
        case .Playing:
            .Playing(turn: (try (try decoder.container(keyedBy: ExtraCodingKeys.self)).decode(CodablePlayerReferenceWrapper.self, forKey: .player)).player)
        case .Win:
            try Self.decodeWin(from: decoder)
        case .Draw:
            .Draw
        case .Invalid:
            .Invalid
        }
    }
    
    private enum ExtraCodingKeys: CodingKey {
        case player, winner, cells
    }
    
    private static func decodeWin(from decoder: any Decoder) throws -> Self {
        guard let context = decoder.userInfo[.gameDecodingContext] as? CodableContext
        else { throw DeError.InvalidVariant }
        
        let container = try decoder.container(keyedBy: ExtraCodingKeys.self)
        
        let winner = (try container.decode(CodablePlayerReferenceWrapper.self, forKey: .winner)).player
        
        let cells = try container.decode([Coords].self, forKey: .cells)
        
        return .Win(winner: winner, board: context.board!, cells: cells)
    }
    
    public func encode(to encoder: any Encoder) throws {
        var container = encoder.container(keyedBy: EnumCommonCodingKeys.self)
        var container2 = encoder.container(keyedBy: ExtraCodingKeys.self)
        
        switch self {
        case .Playing(let turn):
            try container.encode(CodableEnum(of: VariantCodingKeys.Playing), forKey: .type)
            try container2.encode(CodablePlayerReferenceWrapper(of: turn), forKey: .player)
        case .Win(let winner, _, let cells):
            try container.encode(CodableEnum(of: VariantCodingKeys.Win), forKey: .type)
            try container2.encode(CodablePlayerReferenceWrapper(of: winner), forKey: .winner)
            try container2.encode(cells, forKey: .cells)
        case .Draw:
            try container.encode(CodableEnum(of: VariantCodingKeys.Draw), forKey: .type)
        case .Invalid:
            try container.encode(CodableEnum(of: VariantCodingKeys.Invalid), forKey: .type)
        }
    }
}
