import Model

public struct CodablePlayerWrapper: Codable {
    public let player: Player
    
    public init(of player: Player) {
        self.player = player
    }
    
    public init(from decoder: any Decoder) throws {
        let container = try decoder.container(keyedBy: EnumCommonCodingKeys.self)
        
        self.player = try (try container.decode(PlayerType.self, forKey: .type)).decodePlayer(from: decoder)
    }
    
    public func encode(to encoder: any Encoder) throws {
        var container = encoder.container(keyedBy: EnumCommonCodingKeys.self)
        let type = self.player.type
        try container.encode(type, forKey: .type)
        switch type {
        case .Human:
            try CodableHumanPlayerWrapper(of: self.player as! HumanPlayer).encode(to: encoder)
        case .AIRandom:
            try CodableRandomAIPlayerWrapper(of: self.player as! RandomPlayer).encode(to: encoder)
        }
    }
}

extension PlayerType: CodingKey, Codable {
    public init(from decoder: any Decoder) throws {
        self = (try CodableEnum<Self>(from: decoder)).variant
    }
    
    public func encode(to encoder: any Encoder) throws {
        try CodableEnum(of: self).encode(to: encoder)
    }
    
    public func decodePlayer(from decoder: any Decoder) throws -> Player {
        switch self {
        case .Human: (try CodableHumanPlayerWrapper(from: decoder)).player
        case .AIRandom: (try CodableRandomAIPlayerWrapper(from: decoder)).player
        }
    }
}

private enum CommonPlayerCodingKeys: CodingKey {
    case name, pieces
}

private struct CodableHumanPlayerWrapper: Codable {
    let player: HumanPlayer
    
    init(of player: HumanPlayer) {
        self.player = player
    }
    
    init(from decoder: any Decoder) throws {
        guard let context = decoder.userInfo[.gameDecodingContext] as? CodableContext
        else { throw DeError.CodableContext }
        
        let container = try decoder.container(keyedBy: CommonPlayerCodingKeys.self)
        
        let name = try container.decode(String.self, forKey: .name)
        let pieces = try container.decode(CodableEnum<PieceType>.self, forKey: .pieces)
        
        self.init(of: HumanPlayer(name: name, piece_type: pieces.variant, callback: context.callbackFactory!(name, pieces.variant)))
    }
    
    func encode(to encoder: any Encoder) throws {
        var container = encoder.container(keyedBy: CommonPlayerCodingKeys.self)
        
        try container.encode(self.player.name, forKey: .name)
        try container.encode(CodableEnum(of: self.player.piece_type), forKey: .pieces)
    }
}

private struct CodableRandomAIPlayerWrapper: Codable {
    let player: RandomPlayer
    
    init(of player: RandomPlayer) {
        self.player = player
    }
    
    init(from decoder: any Decoder) throws {
        let container = try decoder.container(keyedBy: CommonPlayerCodingKeys.self)
        
        let name = try container.decode(String.self, forKey: .name)
        let pieces = try container.decode(CodableEnum<PieceType>.self, forKey: .pieces)
        
        self.init(of: RandomPlayer(name: name, piece_type: pieces.variant))
    }
    
    func encode(to encoder: any Encoder) throws {
        var container = encoder.container(keyedBy: CommonPlayerCodingKeys.self)
        
        try container.encode(self.player.name, forKey: .name)
        try container.encode(CodableEnum(of: self.player.piece_type), forKey: .pieces)
    }
}

public struct CodablePlayerReferenceWrapper: Codable {
    public let player: Player
    
    public init(of player: Player) {
        self.player = player
    }
    
    public init(from decoder: any Decoder) throws {
        guard
            let context = decoder.userInfo[.gameDecodingContext] as? CodableContext,
            let players = context.players
        else { throw DeError.CodableContext }
        
        let container = try decoder.singleValueContainer()
        
        let id = try container.decode(Int.self)
        guard id < players.count else { throw DeError.InvalidData }
        
        self.player = players[id]
    }
    
    public func encode(to encoder: any Encoder) throws {
        guard
            let context = encoder.userInfo[.gameDecodingContext] as? CodableContext,
            let players = context.players
        else { throw DeError.CodableContext }
        
        var container = encoder.singleValueContainer()
        try container.encode(players.firstIndex(where: { self.player === $0 })!)
    }
}
